/*
 *
 *
 *    RttyModem.ino
 *
 *    A simple terminal/modem for RTTY.
 *
 *    License: GNU General Public License Version 3.0.
 *    
 *    Copyright (C) 2016 by Matthew K. Roberts, KK5JY. All rights reserved.
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *    
 *    This program is distributed in the hope that it will be useful, but
 *    WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    General Public License for more details.
 *    
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see: http://www.gnu.org/licenses/
 *
 *
 */

//
//  Set RTTY_DATA_ANALOG to 'true' to read analog AFSK data from A0.
//  Set to 'false' if using an external demodulator such as the NJM2211.
//
#define RTTY_DATA_ANALOG true

#if RTTY_DATA_ANALOG == true
	//  If reading AFSK data from A0...
	//
	//  Set this to determine the sample rate; divisor should be set
	//  to one of the following values, depending on the sample rate
	//  desired.  The recommended divisor is 128, giving a ~9.6kSps
	//  sampling rate.
	//
	//  Divisor  ->   Sample Rate
	//    128            9615.4
	//     64           19230.8
	//     32           38461.5
	//
	#define RTTY_DATA_SAMPLE_RATE_DIVISOR (128)

	//
	//  Define the zero-volts value to be read from the A/D.  If your
	//  hardware has a perfectly centered DC offset (using a high-
	//  quality voltage divider, etc.), just leave this at 128.  If
	//  you have a larger DC offset, you can change this *slightly*
	//  to compensate.
	//
	#define RTTY_DATA_ANALOG_DC_OFFSET (127)

	//
	//  Set this to 'true' to enable an LED tuning indicator attached
	//  to discrete LEDs.
	//
	#define RTTY_TUNING_INDICATOR true

	// if the tuning indicator is enabled...
	#if RTTY_TUNING_INDICATOR == true
	
	//
	//  Set this to the list of pins used for the dot-graph
	//
	#define RTTY_TUNING_INDICATOR_PINS { 8, 9, 10, 11, 12 }

	//
	//  Set this to the number of LED updates to do per second
	//
	#define RTTY_TUNING_INDICATOR_RATE (5)

	#endif
#else
	//
	//  If using a demodulator circuit such as the NJM2211, set pin
	//  definitions for the receiver.
	//
	//  These are all used for digital I/O, and can be either analog
	//  or digital pins.  Since digital mode is used, use the A#
	//  macros for analog pins (e.g., A0, A1, etc.)
	//

	//
	//  Set the DATA input pin; this is the pin from which logic data
	//  is read from the decoder circuit:
	//
	//  Choice of this pin has limitations, as it uses interrupts; see:
	//     https://www.arduino.cc/en/Reference/AttachInterrupt
	//  ...and pick a pin for your board that supports interrupts.
	#define RTTY_DATA_IN_PIN (3)

	//
	// The DATA CARRIER DETECT (DCD) pin from the PLL (optional); if set,
	// this pin will be read before processing any bits from the decoder
	// hardware; this helps "mute" the receiver when the decoder PLL is
	// unlocked; if set to 0 or -1, the feature will be disabled.
	#define RTTY_DCD_PIN (6)

	//
	// The value to use for active "carrier detect"; reading this logic
	// level from the DCD pin will un-mute the decoding logic; valid
	// values are HIGH and LOW.  If using a 2211 demodulator, the value
	// when using the Q pin is LOW, and the value for the !Q pin is HIGH.
	#define RTTY_DCD_ASSERT (HIGH)
#endif

//
// Set the DATA output pin; this pin can either send logic levels to a
//    transciever's FSK input, or it can generate audio FSK tones to
//    send to a transmitter's audio input:
#define RTTY_DATA_OUT_PIN (5)

//
// Set the PTT output pin; this pin will be driven LOW to activate the
//    radio PTT; set to zero to disable this feature
#define RTTY_PTT_PIN (4)

// Set this to 'true' if the RTTY output line should be logic level,
//   for driving the FSK input of a transciever;
// Set this to 'false' if the output line should emit audio tones:
#define RTTY_DATA_OUT_LOGIC false

// If sending OR receiving AFSK audio data this setting controls the
//    center frequency (in Hz):
#define RTTY_DATA_FREQ_CENTER (1500)

// Likewise, this calculates the shift between tones; one half of
//   this value will be added or subtracted from the CENTER frequency
//   defined above, to generate AFSK output; setting is in Hz:
#define RTTY_DATA_FREQ_SHIFT (170)

// Set the bit rate of the data; standard RTTY is 45.45bps:
#define RTTY_DATA_BIT_RATE (45.45)

// Set this to true to generate and decode inverted RTTY
//   (idle will be higher voltage or frequency)
#define RTTY_INVERTED false

// Heartbeat LED - blinks at 1Hz to show normal loop activity
//    13 == 'L' LED on UNO; set to zero (0) to disable
#define HEARTBEAT_LED (13)

//
//  Host Port - assign a specific hardware port to be the
//              interface to the host; by default, the USB
//              connection on most boards is 'Serial', which
//              is also the hardware TTL serial port on the
//              UNO board.  The hardware serial device on
//              the Leonardo is Serial1.
//
#define HostSerial Serial

// the 'baud rate' of the serial port
#define SerialRate (9600)

// ================== END OF CONFIGURATION SETTINGS ==================

// local headers
#include "CircularBuffer.h"
#include "Elapsed.h"
#define HEARTBEAT_ACTIVE (200000)
#define HEARTBEAT_IN_MICROS
#include "Heartbeat.h"
#include "SplitCommand.h"
#include "baudot.h"
#if RTTY_DATA_ANALOG == true
#include "avraudio.h"
#include "cfsk.h"
#endif
#if RTTY_TUNING_INDICATOR == true
#include "DotGraph.h"
static uint8_t dotGraphPins[] = RTTY_TUNING_INDICATOR_PINS;
static KK5JY::DotGraph dotGraph(dotGraphPins, sizeof(dotGraphPins));
static uint32_t lastTune = 0;
static const uint32_t tuneInterval = 1000000 / RTTY_TUNING_INDICATOR_RATE;
#endif

// the version string
#define VERSION_STRING ("RttyModem 1.0 (beta-20161111a)")

// CR+LF
#define LOCAL_CRLF ("\r\n")

// define RTTY levels
#if RTTY_INVERTED == true
#define RTTY_HIGH LOW
#define RTTY_LOW HIGH
#define RTTY_IDLE HIGH
#else
#define RTTY_HIGH HIGH
#define RTTY_LOW LOW
#define RTTY_IDLE LOW
#endif

// import some namespaces
using namespace KK5JY::Collections;

// define the types of encoding
typedef enum {
	Baudot, // 5N2 ITA-2 Baudot
	ASCII7, // 7N2 ASCII
	ASCII8  // 8N2 ASCII
} Encodings;

//
//  Constants
//

// Limits
const unsigned long MaxPttLeadIn = 1000L * 1000L;	// usec
const unsigned long MaxPttLeadOut = 1000L * 1000L;	// usec
const unsigned long CarrierDetectDelay = 250 * 1000L; // usec
const unsigned long InputStringLength = 16;
const unsigned long CommandStringLength = 16;

// control keys
const char ESC = 0x1B;

////////////////////////////////////////////////////////////////
//
//  Globals
//
////////////////////////////////////////////////////////////////

// the encoding
Encodings encoding = Baudot;

// TX buffer - each element is an encoded byte with start/stop bits added
CircularBuffer<uint16_t> TxBuffer(32);
// TX shift register vars
uint16_t TxMask = 1;

// RX shift register vars
uint16_t RxByte = 0;
uint16_t RxMask = 1;

// a string to hold incoming data
String inputString;

// and another for command data
String cmdText = "";

// command parsing
String command = "";
String argument = "";

// whether the string is complete
bool inputReady = false;

// I/O buffer
char ioBuffer[32];

// echo USB RX data back to host
bool echo = false;

// currently parsing a command
bool cmd = false;

// TX mode (otherwise RX)
bool tx = false;

// tracks PTT state without reading PTT pin
#if RTTY_PTT_PIN > 0
bool pttState = false;
#endif

// tracks the shifted state of the transmitter or receiver
bool txShift = false;
bool txFresh = true;
bool rxShift = false;
bool rxFresh = true;
bool uos = true; // UOS setting for receiver

// current value of RX pin (only assigned by ISR)
volatile uint8_t rttyRxState = false;

// last value of RX pin considered (used for state change detection)
uint8_t rttyRxLastState = LOW;

// RX element timing - stores the last state change time
unsigned long rttyRxEdge = 0;

// RX element timing - stores the last clock edge
unsigned long rttyRxClock = 0;

// TX element timing
unsigned long rttyTxEdge = 0;

// global timer
volatile unsigned long now = 0;

// PTT timing
unsigned long pttLeadIn = 500;	// msec
unsigned long pttLeadOut = 500;	// msec

// DCD timing
unsigned long lastCd = 0;

// holds one bit in the MSB that can be decoded for one byte
uint16_t maxMask = 0;

// holds the stop-bit mask
uint16_t stopBits = 0;

// the bit period (in usec)
uint16_t bitPeriod = 0;

// bit envelope limits
uint32_t maxPeriod = 0;
uint32_t halfBit = 0;
uint32_t roundBit = 0;

#if RTTY_DATA_ANALOG == true
typedef KK5JY::DSP::DelayLineFskDemod Demodulator;
Demodulator *demod = 0;
#endif

////////////////////////////////////////////////////////////////
//
//  Support Methods
//
////////////////////////////////////////////////////////////////

#if RTTY_DATA_ANALOG == true
// stores the audio level for diagnostics
volatile int8_t audio_max = 0;

//
//  A/D Interrupt Handler -- called when new A/D value ready
//
ISR(ADC_vect) {
	// the A/D input var
	static int8_t sample = 0;

	//
	//  NOTE: BE VERY FAST HERE; this handler is being called
	//        thousands of times per second, and doing too
	//        much work here will cause samples to be lost.
	//

	// read the current A/D value and remove the DC offset
	sample = ADCH - RTTY_DATA_ANALOG_DC_OFFSET;

	// check for clipping; store max level
	const uint8_t alevel = abs(sample);
	if (alevel > audio_max)
		audio_max = alevel;

	// do the demod
	rttyRxState = demod->run(sample) ? RTTY_HIGH : RTTY_LOW;
}
#else
//
//  DataInterrupt(...) - ISR for RTTY data input state change
//
static void DataInterrupt() {
	// read the RX line - that's it
	rttyRxState = digitalRead(RTTY_DATA_IN_PIN);
}
#endif


//
//  SetRttyState(bool)
//
static inline void SetRttyState(bool bit) {
#if RTTY_DATA_OUT_LOGIC == true
	// generate logic output
	digitalWrite(RTTY_DATA_OUT_PIN, bit ? RTTY_HIGH : RTTY_LOW);
#else
	// generate AFSK output
	tone(
		RTTY_DATA_OUT_PIN,
		(bit == RTTY_LOW) ?
		RTTY_DATA_FREQ_CENTER + (RTTY_DATA_FREQ_SHIFT / 2) :
		RTTY_DATA_FREQ_CENTER - (RTTY_DATA_FREQ_SHIFT / 2)
	);
#endif

	// record the transition time
	rttyTxEdge = now = micros();
}

//
//  Get the highest mask value that is valid for current encoding
//
static inline uint16_t GetMaxMask() {
	switch(encoding) {
		case Baudot:
			// five data bits, no parity, two stop bits
			//       ssddddds
			return 0b10000000;
		case ASCII7:
			// seven data bits, no parity, two stop bits
			//       ssddddddds
			return 0b1000000000;
		case ASCII8:
			// eight data bits, no parity, two stop bits
			//       ssdddddddds
			return 0b10000000000;
	}
	return 0;
}

//
//  Get the value of the next bit to transmit
//
static inline bool GetNextTxBit(bool & bit) {
	// if there is no data, give up
	uint16_t w;
	if (!TxBuffer.First(w))
		return false;

	// compute the result
	bit = (w & TxMask);

	#if RTTY_INVERTED == true
	if (TxMask != 1 && TxMask != maxMask && TxMask != (maxMask / 2))
		bit = !bit;
	#endif

	// shift the mask to the next bit
	TxMask <<= 1;

	// if we have read all we can get from the current byte...
	if (TxMask > maxMask) {
		// ... go to the next byte
		TxMask = 1;
		TxBuffer.Remove(w);
	}

	// success
	return true;
}


//
//  Decode one byte from the wire
//
static inline bool Decode(uint16_t raw, char &ch) {
	// queue up the character for transmission
	switch (encoding) {
		case Baudot:
			if (raw == RTTY_NUL) {
				ch = 0;
				if (uos) rxShift = false;
				return true;
			} else if (raw == RTTY_CR) {
				ch = '\r';
				if (uos) rxShift = false;
				return true;
			} else if (raw == RTTY_LF) {
				ch = '\n';
				if (uos) rxShift = false;
				return true;
			} else if (raw == RTTY_SP) {
				ch = ' ';
				if (uos) rxShift = false;
				return true;
			} else if (raw == RTTY_LS) {
#ifdef VERBOSE_DEBUG
				HostSerial.write("[LS]");
#endif
				ch = 0; // no printable result
				rxShift = false; // but de-shift the decoder state
				return true; // and return success
			} else if (raw == RTTY_FS) {
#ifdef VERBOSE_DEBUG
				HostSerial.write("[FS]");
#endif

				ch = 0; // no printable result
				rxShift = true; // but up-shift into figures mode
				return true; // and return success
			} else if (rxShift) {
				switch (raw) {
					case RTTY_0: ch = '0'; return true;
					case RTTY_1: ch = '1'; return true;
					case RTTY_2: ch = '2'; return true;
					case RTTY_3: ch = '3'; return true;
					case RTTY_4: ch = '4'; return true;
					case RTTY_5: ch = '5'; return true;
					case RTTY_6: ch = '6'; return true;
					case RTTY_7: ch = '7'; return true;
					case RTTY_8: ch = '8'; return true;
					case RTTY_9: ch = '9'; return true;
					case RTTY_DASH: ch = '-'; return true;
					case RTTY_QUOTE: ch = '\''; return true;
					case RTTY_DOLLAR: ch = '$'; return true;
					case RTTY_BANG: ch = '!'; return true;
					case RTTY_AMP: ch = '&'; return true;
					case RTTY_HASH: ch = '#'; return true;
					case RTTY_PARL: ch = '('; return true;
					case RTTY_PARR: ch = ')'; return true;
					case RTTY_PLUS: ch = '+'; return true;
					case RTTY_SLASH: ch = '/'; return true;
					case RTTY_SEMI: ch = ';'; return true;
					case RTTY_QUES: ch = '?'; return true;
					case RTTY_COMMA: ch = ','; return true;
					case RTTY_FSTOP: ch = '.'; return true;
					case RTTY_COLON: ch = ':'; return true;
					default: break;
				}
			} else { // unshifted
				switch (raw) {
					case RTTY_A: ch = 'A'; return true;
					case RTTY_B: ch = 'B'; return true;
					case RTTY_C: ch = 'C'; return true;
					case RTTY_D: ch = 'D'; return true;
					case RTTY_E: ch = 'E'; return true;
					case RTTY_F: ch = 'F'; return true;
					case RTTY_G: ch = 'G'; return true;
					case RTTY_H: ch = 'H'; return true;
					case RTTY_I: ch = 'I'; return true;
					case RTTY_J: ch = 'J'; return true;
					case RTTY_K: ch = 'K'; return true;
					case RTTY_L: ch = 'L'; return true;
					case RTTY_M: ch = 'M'; return true;
					case RTTY_N: ch = 'N'; return true;
					case RTTY_O: ch = 'O'; return true;
					case RTTY_P: ch = 'P'; return true;
					case RTTY_Q: ch = 'Q'; return true;
					case RTTY_R: ch = 'R'; return true;
					case RTTY_S: ch = 'S'; return true;
					case RTTY_T: ch = 'T'; return true;
					case RTTY_U: ch = 'U'; return true;
					case RTTY_V: ch = 'V'; return true;
					case RTTY_W: ch = 'W'; return true;
					case RTTY_X: ch = 'X'; return true;
					case RTTY_Y: ch = 'Y'; return true;
					case RTTY_Z: ch = 'Z'; return true;
					default: break;
				}
			}

#ifdef VERBOSE_DEBUG
			HostSerial.print("#debug,Could not decode symbol: ");
			HostSerial.print(raw);
			HostSerial.println(";");
#endif
			return false;

		case ASCII7:
			ch = (raw >> 1);
			return true;
		case ASCII8:
			ch = (raw >> 1);
			return true;
		default:
			return false;
	}
}

//
//  Enqueue the next byte for transmission
//
static inline int8_t Encode(char ch) {
	// queue up the character for transmission
	switch (encoding) {
		case Baudot:
			// start with the characters that are shift-independent
			if (ch == 0) {
				return TxBuffer.Add((uint16_t)(RTTY_NUL)) ? 1 : -1;
			} else if (ch == '\r') {
				return TxBuffer.Add((uint16_t)(RTTY_CR)) ? 1 : -1;
			} else if (ch == '\n') {
				return TxBuffer.Add((uint16_t)(RTTY_LF)) ? 1 : -1;
			} else if (ch == ' ') {
				return TxBuffer.Add((uint16_t)(RTTY_SP)) ? 1 : -1;
			} else if (isdigit(ch)) {
				int8_t bc = 0;

				// shift
				if (txFresh || !txShift) {
					TxBuffer.Add((uint16_t)(RTTY_FS));
					txShift = true;
					txFresh = false;
					bc = 1;
				}

				// transform the digit
				switch (ch) {
					case '0': ch = RTTY_0; break;
					case '1': ch = RTTY_1; break;
					case '2': ch = RTTY_2; break;
					case '3': ch = RTTY_3; break;
					case '4': ch = RTTY_4; break;
					case '5': ch = RTTY_5; break;
					case '6': ch = RTTY_6; break;
					case '7': ch = RTTY_7; break;
					case '8': ch = RTTY_8; break;
					case '9': ch = RTTY_9; break;
					default: return 0;
				}

				return TxBuffer.Add((uint16_t)(ch)) ? bc + 1 : -1; // this catches overflow from the FS insertion, too

			} else if (isalpha(ch)) {
				int8_t bc = 0;

				// shift
				if (txShift || txFresh) {
					TxBuffer.Add((uint16_t)(RTTY_LS));
					txShift = false;
					txFresh = false;
					bc = 1;
				}

				// transform the character
				switch (toupper(ch)) {
					case 'A': ch = RTTY_A; break;
					case 'B': ch = RTTY_B; break;
					case 'C': ch = RTTY_C; break;
					case 'D': ch = RTTY_D; break;
					case 'E': ch = RTTY_E; break;
					case 'F': ch = RTTY_F; break;
					case 'G': ch = RTTY_G; break;
					case 'H': ch = RTTY_H; break;
					case 'I': ch = RTTY_I; break;
					case 'J': ch = RTTY_J; break;
					case 'K': ch = RTTY_K; break;
					case 'L': ch = RTTY_L; break;
					case 'M': ch = RTTY_M; break;
					case 'N': ch = RTTY_N; break;
					case 'O': ch = RTTY_O; break;
					case 'P': ch = RTTY_P; break;
					case 'Q': ch = RTTY_Q; break;
					case 'R': ch = RTTY_R; break;
					case 'S': ch = RTTY_S; break;
					case 'T': ch = RTTY_T; break;
					case 'U': ch = RTTY_U; break;
					case 'V': ch = RTTY_V; break;
					case 'W': ch = RTTY_W; break;
					case 'X': ch = RTTY_X; break;
					case 'Y': ch = RTTY_Y; break;
					case 'Z': ch = RTTY_Z; break;
					default: return 0;
				}

				return TxBuffer.Add((uint16_t)(ch)) ? bc + 1 : -1; // this catches overflow from the LS insertion, too

			} else { // some of the symbols
				switch (ch) {
					case '-': ch = RTTY_DASH; break;
					case '\'': ch = RTTY_QUOTE; break;
					case '$': ch = RTTY_DOLLAR; break;
					case '!': ch = RTTY_BANG; break;
					case '&': ch = RTTY_AMP; break;
					case '#': ch = RTTY_HASH; break;
					case '(': ch = RTTY_PARL; break;
					case ')': ch = RTTY_PARR; break;
					case '+': ch = RTTY_PLUS; break;
					case '/': ch = RTTY_SLASH; break;
					case ';': ch = RTTY_SEMI; break;
					case ':': ch = RTTY_COLON; break;
					case '?': ch = RTTY_QUES; break;
					case ',': ch = RTTY_COMMA; break;
					case '.': ch = RTTY_FSTOP; break;
					default: ch = 0; break;
				}
				if (!ch) {
#ifdef VERBOSE_DEBUG
					HostSerial.print("#debug: couldn't encode symbol;");
#endif
					// give up
					return 0;
				}

				int8_t bc = 0;
				if (txFresh || !txShift) {
					TxBuffer.Add((uint16_t)(RTTY_FS));
					txShift = true;
					txFresh = false;
					bc = 1;
				}

				return TxBuffer.Add((uint16_t)(ch)) ? bc + 1 : -1; // this catches overflow from the FS insertion, too
			}
			return false;
		case ASCII7:
			return TxBuffer.Add((uint16_t)(((ch & 0x7F) << 1) & 0x0001)) ? 1 : -1;
		case ASCII8:
			return TxBuffer.Add((uint16_t)((ch << 1) & 0x0001)) ? 1 : -1;
		default:
			return false;
	}
}


//
//  ProcessShiftRegister() - common code for both reciever's shift register cycle
//
static inline bool ProcessShiftRegister(uint8_t bit) {
	bool result = false;

	// if the input is HIGH, set the LSB
	if (bit == RTTY_HIGH) {
		RxByte |= RxMask;
	}

	// left-shift to next position
	if (RxMask < maxMask) {
		RxMask <<= 1;
	} else {
		//
		// check for decoded byte
		//
		// at this point, we should have one complete byte,
		//    stop/start bits and all;
		// if so, decode;
		// if not, shift to the first candidate start bit and move on
		//
		if ((RxByte & 0x0001) && ((RxByte & stopBits) == 0x0000)) {
			// valid byte!  Decode and print...
			char ch;
			if (result = Decode(RxByte, ch)) {
				// if the character can be printed, print it
				switch (ch) {
					case 0: break;
					case '#': {
						HostSerial.print("#HASH;");
					} break;
					default: {
						HostSerial.print(ch);
					} break;
				}

				// reset shift register
				RxByte = 0;
				RxMask = 1;
			}
		}

		// if the shift register wasn't cleared by a successful decode,
		//    shift it over by one bit to allow the contents to be
		//    reused in successive attempts
		if (RxByte)
			RxByte >>= 1;
	}

	return result;
}


////////////////////////////////////////////////////////////////
//
//  MAIN PROGRAM
//
////////////////////////////////////////////////////////////////

//
//  setup() - initialize hardware
//
void setup() {
	// start the serial port
	HostSerial.begin(SerialRate);
	// and wait for it to initialize
	while (!HostSerial) { /* nop */ }

	// allocate strings
	cmdText.reserve(CommandStringLength);
	inputString.reserve(InputStringLength);
	inputString = strcpy_P(ioBuffer, PSTR("#VERSION;"));
	inputReady = true;

	// RTTY PTT
#if RTTY_PTT_PIN > 0
	pinMode(RTTY_PTT_PIN, OUTPUT);
	digitalWrite(RTTY_PTT_PIN, LOW);
	pttState = false;
#endif

	// HB LED
#if HEARTBEAT_LED > 0
	HeartbeatSetup();
#endif

	// compute the mask value that causes a new byte to be used
	maxMask = GetMaxMask();
	// compute the stop-bit mask
	stopBits = (maxMask >> 1) | maxMask;
	// compute the bit period (usec)
	bitPeriod = round(1000000.0 / RTTY_DATA_BIT_RATE);
	// calculate the max bit envelope
	maxPeriod = 16L * bitPeriod;
	// calculate the smallest bit time
	halfBit = bitPeriod / 2L;
	// calculate the length, above a whole bit length, that results in one additional bit
	roundBit = ((1 * bitPeriod) / 5); // TODO: not sure why this is 1/5...

	#if RTTY_DATA_ANALOG == true
	//
	//  configure the AFSK demodulator
	//
	{
		double corr = 0.0;
		switch (RTTY_DATA_SAMPLE_RATE_DIVISOR) {
			case 128: corr = -0.011; break;
			case 64: corr = -0.001; break;
		}

		// config
		Demodulator::Config dc;
		dc.f0 = RTTY_DATA_FREQ_CENTER;
		dc.corr = corr;
		dc.fs = GetFs(RTTY_DATA_SAMPLE_RATE_DIVISOR);
		dc.int_len = (dc.fs / (2.0 * RTTY_DATA_BIT_RATE)); // 1/2 bit time
		dc.trk_len = dc.int_len * 16; // a few bits, to make sure to see both dev rails
		dc.ctr_limit = 10;
		dc.ctr_trigger = 5;

		// instantiate
		demod = new Demodulator(dc);
	}
	#else
	// RTTY DATA IN: enable digital input for DATA
	pinMode(RTTY_DATA_IN_PIN, INPUT_PULLUP);
	// RTTY DATA OUT: enable digital output for DATA
	pinMode(RTTY_DATA_OUT_PIN, OUTPUT);
	digitalWrite(RTTY_DATA_OUT_PIN, RTTY_IDLE);
	
	// RTTY DCD: enable digital input on DCD pin
	#if RTTY_DCD_PIN > 0
	pinMode(RTTY_DCD_PIN, INPUT_PULLUP);
	#endif

	// RTTY RX DATA: attach interrupt to input pin
	attachInterrupt(digitalPinToInterrupt(RTTY_DATA_IN_PIN), DataInterrupt, CHANGE);
	#endif

	#if RTTY_DATA_ANALOG == true
	// setup the A/D input
	ConfigureAnalogSampling(RTTY_DATA_SAMPLE_RATE_DIVISOR);
	#endif

	#if RTTY_TUNING_INDICATOR == true
	dotGraph.test();
	#endif
}


//
//  loop() - main program loop
//
void loop() {
	static bool hash = false;

	// read the clock
	now = micros();

	// blink the HB LED
	HeartbeatLoop();

	// process PTT key-up delay
	if (!tx &&
		#if RTTY_PTT_PIN > 0
		pttState &&
		#endif
		Elapsed(now, rttyTxEdge) >= (pttLeadOut * 1000)) {
		#if RTTY_PTT_PIN > 0
		digitalWrite(RTTY_PTT_PIN, LOW);
		pttState = false;
		#endif
		#if RTTY_DATA_OUT_LOGIC == false // if AFSK TX...
		noTone(RTTY_DATA_OUT_PIN); // disable the tone output
		#endif
		#if RTTY_DATA_ANALOG == true // if AFSK RX...
		SetSamplingStatus(true); // re-enable the receiver sampling
		#endif
	}
	
	// if we are in transmit mode, service the TX queue
	if (tx) {
		// re-read the clock
		now = micros();

		// if the current element has been sent...
		if (Elapsed(now, rttyTxEdge) >= bitPeriod) {
			bool bit;
			if (GetNextTxBit(bit)) {
				// generate the leading edge
				SetRttyState(bit);
			} else {
				tx = false;
				SetRttyState(RTTY_IDLE);
			}
		}
	} // if (tx)...

	// if user input ready
	if (inputReady || hash) {
		// for each char received from the USB...
		for (int i = 0; hash || (i != inputString.length()); i += hash ? 0 : 1) {
			char ch = hash ? '#' : inputString.charAt(i);
			
			// if we are in command mode...
			if (cmd) {
				if (ch == '#') {
					// if a '#' sent mid-command, start over
					cmdText = "";
				} else if (ch == ';') {
					// terminate and process the command string
					cmd = false;
					bool ok = false;
					if (echo) HostSerial.print(';');

					// parse the command and argument
					SplitCommand(cmdText, command, argument);
					if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("DOC")))) {
						//
						//  QUERY DOC
						//
						if (argument.length() == 0) {
							ok = true;
							cmdText += strcpy_P(ioBuffer, PSTR(":UOS,ECHO,PTTIN,"));
							cmdText += strcpy_P(ioBuffer, PSTR("TUNE,PTTOUT,VERSION"));
						} else {
							//
							//  TODO: respond to individual doc requests
							//
						}
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("UOS")))) {
						//
						//  QUERY/SET RX UOS
						//
						if (argument.length() == 0) {
							ok = true;
							cmdText += '=';
							cmdText += uos ? "1" : "0";
						} else {
							int newUos = argument.toInt();
							uos = newUos ? true : false;
							ok = true;
						}
#if RTTY_DATA_ANALOG == true
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("LEVELMAX")))) {
						// return the highest volume level since command last run
						ok = true;
						cmdText += '=';
						cmdText += 128;
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("LEVEL")))) {
						// return the highest volume level since command last run
						ok = true;
						cmdText += '=';
						cmdText += audio_max;
						audio_max = 0;
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("DEV")))) {
						// return the demod's current deviation value (mostly for testing, alignment)
						ok = true;
						cmdText += '=';
						cmdText += demod->deviation();
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("CTR")))) {
						// return the demod's current output counter value (mostly for testing, alignment)
						ok = true;
						cmdText += '=';
						cmdText += demod->counter();
#endif
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("DS")))) {
						// return the current input data sate
						ok = true;
						cmdText += '=';
						cmdText += rttyRxState;
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("ECHO")))) {
						//
						//  QUERY/SET ECHO
						//
						if (argument.length() == 0) {
							ok = true;
							cmdText += '=';
							cmdText += echo ? "1" : "0";
						} else {
							int newEcho = argument.toInt();
							echo = newEcho ? true : false;
							ok = true;
						}
#if RTTY_DATA_OUT_LOGIC == false
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("TUNE")))) {
						//
						//  SET TUNE
						//
						if ((argument.length() == 1) && !tx) {
							char doTune = argument.charAt(0);
							switch (doTune) {
								case '0':
									// mute the tone
									noTone(RTTY_DATA_OUT_PIN);

									#if RTTY_PTT_PIN > 0
									// de-PTT the radio
									digitalWrite(RTTY_PTT_PIN, LOW);
									pttState = false;
									#endif
									#if RTTY_DATA_ANALOG == true
									SetSamplingStatus(true); // re-enable the receiver sampling
									#endif

									ok = true;
									break;
								case '1':
									// emit tone at center freq
									tone(RTTY_DATA_OUT_PIN, RTTY_DATA_FREQ_CENTER);

									#if RTTY_DATA_ANALOG == true
									SetSamplingStatus(false); // disable the receiver sampling
									#endif
									#if RTTY_PTT_PIN > 0
									// PTT the radio
									digitalWrite(RTTY_PTT_PIN, HIGH);
									pttState = false;
									#endif

									ok = true;
									break;
								default:
									// nop
									break;
							}
						}
#endif
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("PTTIN")))) {
						//
						//  QUERY/SET PTT LEAD-IN
						//
						if (argument.length() == 0) {
							ok = true;
							cmdText += '=';
							cmdText += pttLeadIn;
						} else {
							unsigned long newPTT = argument.toInt();
							if (newPTT >= 0 && newPTT <= MaxPttLeadIn) {
								pttLeadIn = newPTT;
								ok = true;
							}
						}
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("PTTOUT")))) {
						//
						//  QUERY/SET PTT LEAD-OUT
						//
						if (argument.length() == 0) {
							ok = true;
							cmdText += '=';
							cmdText += pttLeadOut;
						} else {
							unsigned long newPTT = argument.toInt();
							if (newPTT >= 0 && newPTT <= MaxPttLeadOut) {
								pttLeadOut = newPTT;
								ok = true;
							}
						}
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("VERSION")))) {
						//
						//  QUERY VERSION NUMBER
						//
						ok = true;
						cmdText += '=';
						cmdText += strcpy_P(ioBuffer, PSTR(VERSION_STRING));
					} else if (command.equalsIgnoreCase(strcpy_P(ioBuffer, PSTR("HASH")))) {
						//
						//  SEND HASH
						//
						hash = true;
						ok = true;
					}
					
					//
					//  TODO: process other command contents
					//

					// respond with #OK or #ERR
					if (echo)
						HostSerial.print(strcpy_P(ioBuffer, PSTR(LOCAL_CRLF)));
					HostSerial.print(ok ? strcpy_P(ioBuffer, PSTR("#OK:")) : strcpy_P(ioBuffer, PSTR("#ERR:")));
					HostSerial.print(cmdText);
					HostSerial.print(';');
					if (echo)
						HostSerial.print(strcpy_P(ioBuffer, PSTR(LOCAL_CRLF)));
					cmdText = "";
				} else {
					if (isAlpha(ch) || isDigit(ch) || isPunct(ch) || (ch == ' ')) {
						// add the char to the command string
						if (cmdText.length() < CommandStringLength)
							cmdText += ch;
						
						// echo the char, if enabled
						if (echo) {
							HostSerial.print(ch);
						}
					}
				}
				continue;
			}
			
			// start command mode??
			if (!hash) {
				if ((!cmd) && (ch == '#')) {
					cmd = true;
					if (echo) HostSerial.print('#');
					continue;
				}
			}

			// hard-stop?
			if (ch == ESC) {
				#if RTTY_PTT_PIN > 0
				digitalWrite(RTTY_PTT_PIN, LOW);
				pttState = false;
				#endif
				TxBuffer.Clear();
				tx = false;
				#if RTTY_DATA_OUT_LOGIC == false // if AFSK TX...
				noTone(RTTY_DATA_OUT_PIN); // disable the tone output
				#endif
				#if RTTY_DATA_ANALOG == true
				SetSamplingStatus(true); // re-enable the receiver sampling
				#endif
				continue;
			}

			// echo the char, if enabled
			if (hash) {
				hash = false;
			} else if (echo) {
				HostSerial.print(ch);
			}

			//
			//  TX - drive the output pin and PTT
			//
			int8_t eok = Encode(ch);
			if (eok < 0) {
				HostSerial.print(strcpy_P(ioBuffer, PSTR("#OVF;")));
			}
			if (eok && !tx) {
				bool bit;
				if (GetNextTxBit(bit)) {
#if RTTY_DATA_ANALOG == true
					SetSamplingStatus(false); // disable the receiver sampling
#endif
					// generate idle
					SetRttyState(RTTY_IDLE);

					// PTT lead-in
					#if RTTY_PTT_PIN > 0
					pttState = true;
					digitalWrite(RTTY_PTT_PIN, HIGH);
					#endif
					delay(pttLeadIn);

					// start sending
					tx = true;
					txFresh = true;

					// generate the leading edge
					SetRttyState(bit);
				}
			} // Encode(...) && !tx
		}

done:
		// clear the string:
		inputString = "";
		inputReady = false;
	}

	//
	//  RX - process incoming bits
	//
	if (!tx) { // if not transmitting...
		const uint8_t newState = rttyRxState;

		now = micros();
		uint32_t bitLength = Elapsed(now, rttyRxEdge);
		if (bitLength >= (maxPeriod)) {
			bitLength = maxPeriod;
		}

		// if state changed enough for a complete bit
		//    OR if the line is idle for more than a full byte-time....
		const bool stateChanged = (newState != rttyRxLastState);
		const bool fullBit = (bitLength >= halfBit);
		const bool tooIdle = (bitLength >= maxPeriod);
		const bool srActive = (RxMask != 1);
		bool decoded = false;
		if ((stateChanged && fullBit) || (tooIdle && srActive)) {
			const uint8_t currentState = rttyRxLastState;	// save the previous state
			rttyRxLastState = newState;						// save the new state (for later edge triggering)

			// save this edge time
			rttyRxEdge = now;

#if RTTY_DCD_PIN > 0
			// evaluate the current DCD state from one of the 'Q' pins from the PLL
			bool deSql = (lastCd != 0) && (Elapsed(now, lastCd) < CarrierDetectDelay);
			uint8_t dcd = digitalRead(RTTY_DCD_PIN);
			if (dcd == RTTY_DCD_ASSERT) {
				lastCd = now;
				if (!lastCd) {
					lastCd = 1;
				}
				deSql = true;
			}

			// only process bits if CD is active
			if (deSql) {
#endif
				// at this point we have (length / period) bits of value 'currentState' bits,
				//    rounded to the nearest whole bit
				uint32_t bits = (bitLength / bitPeriod);
				if ((bitLength % bitPeriod) >= roundBit) {
					++bits;
				}

				// for each bit
				for (uint32_t i = 0; i != bits; ++i) {
					// add the bit and process the contents of the shift register
					decoded = ProcessShiftRegister(currentState == RTTY_HIGH);
				}
#if RTTY_DCD_PIN > 0
			} else {
				lastCd = 0;
			}
#endif
		}

		//
		//  TUNING INDICATOR
		//
		#if RTTY_TUNING_INDICATOR == true
		static bool moving = false;
		moving |= decoded;
		if (Elapsed(now, lastTune) >= tuneInterval) {
			lastTune = now;

			// read the current deviation
			int16_t dev = demod->deviation();

			// clamp it to int-8 space
			if (dev >= 127) {
				dev = 127;
			} else if (dev <= -128) {
				dev = -128;
			}

			// and update the bar graph
			if (moving)
				dotGraph.update(dev);
			else
				dotGraph.clear();

			moving = false;
		}
		#endif
	}

	//
	//  HOST SERIAL/USB - service the serial port RX buffer
	//
	while (HostSerial.available() && (inputString.length() < InputStringLength)) {
		// get the new byte:
		char inChar = (char)HostSerial.read();
		// add it to the inputString:
		inputString += inChar;
		inputReady = true;
	}
}

// EOF
